/*
 * Copyright (c) 2006 Oracle.  All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
#include <linux/kernel.h>
#include <linux/list.h>
#include <net/inet_hashtables.h>

#include "rds.h"
#include "loop.h"

#define RDS_CONNECTION_HASH_BITS 12
#define RDS_CONNECTION_HASH_ENTRIES (1 << RDS_CONNECTION_HASH_BITS)
#define RDS_CONNECTION_HASH_MASK (RDS_CONNECTION_HASH_ENTRIES - 1)

/* converting this to RCU is a chore for another day.. */
static spinlock_t rds_conn_lock = SPIN_LOCK_UNLOCKED;
static unsigned long rds_conn_count;
struct hlist_head rds_conn_hash[RDS_CONNECTION_HASH_ENTRIES];
static struct kmem_cache *rds_conn_slab;

static struct hlist_head *rds_conn_bucket(__be32 laddr, __be32 faddr)
{
	unsigned long hash = inet_ehashfn(be32_to_cpu(laddr), 0, 
					  be32_to_cpu(faddr), 0);
	return &rds_conn_hash[hash & RDS_CONNECTION_HASH_MASK];
}

#define rds_conn_info_set(var, test, suffix) do {		\
	if (test)						\
		var |= RDS_INFO_CONNECTION_FLAG_##suffix;	\
} while (0)

static inline int rds_conn_is_sending(struct rds_connection *conn)
{
	int ret = 0;

	if (down_trylock(&conn->c_send_sem))
		ret = 1;
	else
		up(&conn->c_send_sem);

	return ret;
}

static struct rds_connection *rds_conn_lookup(struct hlist_head *head,
					      __be32 laddr, __be32 faddr)
{
	struct rds_connection *conn, *ret = NULL;
	struct hlist_node *pos;

	hlist_for_each_entry(conn, pos, head, c_hash_node) {
		if (conn->c_faddr == faddr && conn->c_laddr == laddr) {
			ret = conn;
			break;
		}
	}
	rdsdebug("returning conn %p for %u.%u.%u.%u -> %u.%u.%u.%u\n", ret,
		 NIPQUAD(laddr), NIPQUAD(faddr));
	return ret;
}

/*
 * This is called by transports as they're bringing down a connection.
 * It clears partial message state so that the transport can start sending
 * and receiving over this connection again in the future.  It is up to
 * the transport to have serialized this call with its send and recv.
 */
void rds_conn_reset(struct rds_connection *conn)
{
	unsigned long flags;

	rds_stats_inc(s_conn_reset);

	spin_lock_irqsave(&conn->c_lock, flags);

	rds_send_reset(conn);
	list_splice_init(&conn->c_retrans, &conn->c_send_queue);

	spin_unlock_irqrestore(&conn->c_lock, flags);

	conn->c_next_rx_seq = 0;
}

/*
 * There is only every one 'conn' for a given pair of addresses in the
 * system at a time.  They contain messages to be retransmitted and so
 * span the lifetime of the actual underlying transport connections.
 *
 * For now they are not garbage collected once they're created.  They
 * are torn down as the module is removed, if ever.
 */
struct rds_connection *rds_conn_create(__be32 laddr, __be32 faddr,
				       struct rds_transport *trans, gfp_t gfp)
{
	struct rds_connection *conn, *tmp;
	struct hlist_head *head = rds_conn_bucket(laddr, faddr);
	unsigned long flags;
	int ret;

	spin_lock_irqsave(&rds_conn_lock, flags);
	conn = rds_conn_lookup(head, laddr, faddr);
	spin_unlock_irqrestore(&rds_conn_lock, flags);
	if (conn)
		goto out;

	conn = kmem_cache_alloc(rds_conn_slab, gfp);
	if (conn == NULL) {
		conn = ERR_PTR(-ENOMEM);
		goto out;
	}

	/* hash_node below */
	conn->c_laddr = laddr;
	conn->c_faddr = faddr;
	spin_lock_init(&conn->c_lock);
	conn->c_next_tx_seq = 0;

	init_MUTEX(&conn->c_send_sem);
	conn->c_xmit_rm = NULL;
	conn->c_xmit_sg = 0;
	conn->c_xmit_hdr_off = 0;
	conn->c_xmit_data_off = 0;

	INIT_LIST_HEAD(&conn->c_send_queue);
	INIT_LIST_HEAD(&conn->c_retrans);

	conn->c_next_rx_seq = 0;

	conn->c_map_queued = 0;
	conn->c_map_offset = 0;
	conn->c_map_bytes = 0;

	ret = rds_cong_get_maps(conn);
	if (ret) {
		kmem_cache_free(rds_conn_slab, conn);
		conn = ERR_PTR(ret);
		goto out;
	}

	/*
	 * This is where a connection becomes loopback.  If *any* RDS sockets
	 * can bind to the destination address then we'd rather the messages
	 * flow through loopback rather than either transport.
	 */
	if (rds_trans_get_preferred(faddr))
		trans = &rds_loop_transport;

	conn->c_trans = trans;

	ret = trans->conn_alloc(conn, gfp);
	if (ret) {
		kmem_cache_free(rds_conn_slab, conn);
		conn = ERR_PTR(ret);
		goto out;
	}

	clear_bit(RDS_CONN_CONNECTING, &conn->c_status);
	clear_bit(RDS_CONN_CONNECTED, &conn->c_status);
	conn->c_reconnect_jiffies = 0;
	INIT_DELAYED_WORK(&conn->c_send_w, rds_send_worker);
	INIT_DELAYED_WORK(&conn->c_recv_w, rds_recv_worker);
	INIT_DELAYED_WORK(&conn->c_conn_w, rds_connect_worker);
	INIT_WORK(&conn->c_down_w, rds_shutdown_worker);

	rdsdebug("allocated conn %p for %u.%u.%u.%u -> %u.%u.%u.%u\n", conn,
		 NIPQUAD(laddr), NIPQUAD(faddr));

	spin_lock_irqsave(&rds_conn_lock, flags);
	tmp = rds_conn_lookup(head, laddr, faddr);
	if (tmp) {
		trans->conn_free(conn->c_transport_data);
		kmem_cache_free(rds_conn_slab, conn);
		conn = tmp;
	} else {
		rds_cong_add_conn(conn);
		hlist_add_head(&conn->c_hash_node, head);
		rds_conn_count++;
	}

	spin_unlock_irqrestore(&rds_conn_lock, flags);

out:
	return conn;
}
EXPORT_SYMBOL_GPL(rds_conn_create);

static void rds_conn_message_info(struct socket *sock, unsigned int len,
			          struct rds_info_iterator *iter,
			          struct rds_info_lengths *lens,
			          int want_send)
{
	struct hlist_head *head;
	struct hlist_node *pos;
	struct list_head *list;
	struct rds_connection *conn;
	struct rds_message *rm;
	unsigned long flags;
	unsigned int total = 0;
	size_t i;

	len /= sizeof(struct rds_info_message);

	spin_lock_irqsave(&rds_conn_lock, flags);

	for (i = 0, head = rds_conn_hash; i < ARRAY_SIZE(rds_conn_hash);
	     i++, head++) {
		hlist_for_each_entry(conn, pos, head, c_hash_node) {
			if (want_send)
				list = &conn->c_send_queue;
			else
				list = &conn->c_retrans;

			spin_lock(&conn->c_lock);

			/* XXX too lazy to maintain counts.. */
			list_for_each_entry(rm, list, m_conn_item) {
				total++;
				if (total <= len)
					rds_inc_info_copy(&rm->m_inc, iter,
							  conn->c_laddr,
							  conn->c_faddr, 0);
			}

			spin_unlock(&conn->c_lock);
		}
	}

	spin_unlock_irqrestore(&rds_conn_lock, flags);

	lens->nr = total;
	lens->each = sizeof(struct rds_info_message);
}

static void rds_conn_message_info_send(struct socket *sock, unsigned int len,
				       struct rds_info_iterator *iter,
				       struct rds_info_lengths *lens)
{
	rds_conn_message_info(sock, len, iter, lens, 1);
}

static void rds_conn_message_info_retrans(struct socket *sock,
					  unsigned int len,
					  struct rds_info_iterator *iter,
					  struct rds_info_lengths *lens)
{
	rds_conn_message_info(sock, len, iter, lens, 0);
}

static void rds_conn_info(struct socket *sock, unsigned int len,
			  struct rds_info_iterator *iter,
			  struct rds_info_lengths *lens)
{
	struct rds_info_connection cinfo;
	struct hlist_head *head;
	struct hlist_node *pos;
	struct hlist_node *tmp;
	struct rds_connection *conn;
	unsigned long flags;
	size_t i;

	spin_lock_irqsave(&rds_conn_lock, flags);

	if (len / sizeof(cinfo) < rds_conn_count)
		goto out;

	for (i = 0, head = rds_conn_hash; i < ARRAY_SIZE(rds_conn_hash);
	     i++, head++) {
		hlist_for_each_entry_safe(conn, pos, tmp, head, c_hash_node) {

			/* XXX no c_lock usage.. */

			cinfo.next_tx_seq = conn->c_next_tx_seq;
			cinfo.next_rx_seq = conn->c_next_rx_seq;
			cinfo.laddr = conn->c_laddr;
			cinfo.faddr = conn->c_faddr;
			strncpy(cinfo.transport, conn->c_trans->t_name,
				sizeof(cinfo.transport));
			cinfo.flags = 0;

			rds_conn_info_set(cinfo.flags,
					  rds_conn_is_sending(conn), SENDING);
			rds_conn_info_set(cinfo.flags,
					  test_bit(RDS_CONN_CONNECTING,
						   &conn->c_status),
					  CONNECTING);
			rds_conn_info_set(cinfo.flags,
					  test_bit(RDS_CONN_CONNECTED,
						   &conn->c_status),
					  CONNECTED);

			rds_info_copy(iter, &cinfo, sizeof(cinfo));
		}
	}

out:
	lens->nr = rds_conn_count;
	lens->each = sizeof(cinfo);

	spin_unlock_irqrestore(&rds_conn_lock, flags);
}

int __init rds_conn_init(void)
{
	rds_conn_slab = kmem_cache_create("rds_connection",
					  sizeof(struct rds_connection),
				          0, 0, NULL, NULL);
	if (rds_conn_slab == NULL)
		return -ENOMEM;

	rds_info_register_func(RDS_INFO_CONNECTIONS, rds_conn_info);
	rds_info_register_func(RDS_INFO_SEND_MESSAGES,
			       rds_conn_message_info_send);
	rds_info_register_func(RDS_INFO_RETRANS_MESSAGES,
			       rds_conn_message_info_retrans);

	return 0;
}

void __exit rds_conn_exit(void)
{
	struct hlist_head *head;
	struct hlist_node *pos, *tmp;
	struct rds_connection *conn;
	struct rds_message *rm, *rtmp;
	size_t i;

	for (i = 0, head = rds_conn_hash; i < ARRAY_SIZE(rds_conn_hash);
	     i++, head++) {
		hlist_for_each_entry_safe(conn, pos, tmp, head, c_hash_node) {

			/* the conn won't reconnect once it's unhashed */
			hlist_del_init(&conn->c_hash_node);
			rds_conn_count--;

			rdsdebug("freeing conn %p for %u.%u.%u.%u -> "
				 "%u.%u.%u.%u\n", conn, NIPQUAD(conn->c_laddr),
				 NIPQUAD(conn->c_faddr));

			/* wait for the rds thread to shut it down */
			cancel_delayed_work(&conn->c_conn_w);
			queue_work(rds_wq, &conn->c_down_w);
			flush_workqueue(rds_wq);

			/* tear down queued messages */
			list_for_each_entry_safe(rm, rtmp,
						 &conn->c_send_queue,
						 m_conn_item) {
				list_del_init(&rm->m_conn_item);
				BUG_ON(!list_empty(&rm->m_sock_item));
				rds_message_put(rm);
			}
			if (conn->c_xmit_rm)
				rds_message_put(conn->c_xmit_rm);

			conn->c_trans->conn_free(conn->c_transport_data);

			/*
			 * The congestion maps aren't freed up here.  They're
			 * freed by rds_cong_exit() after all the connections
			 * have been freed.
			 */
			rds_cong_remove_conn(conn);

			BUG_ON(!list_empty(&conn->c_retrans));
			kmem_cache_free(rds_conn_slab, conn);
		}
	}

	kmem_cache_destroy(rds_conn_slab);

	rds_info_deregister_func(RDS_INFO_CONNECTIONS, rds_conn_info);
	rds_info_deregister_func(RDS_INFO_SEND_MESSAGES,
				 rds_conn_message_info_send);
	rds_info_deregister_func(RDS_INFO_RETRANS_MESSAGES,
				 rds_conn_message_info_retrans);
}
